"""The twill multiprocess execution system."""

import os
import sys
import time
from argparse import ArgumentParser
from pickle import dump, load

from twill import execute_file, set_log_level


def main() -> None:
    """Run twill scripts in parallel."""
    try:
        if sys.platform == "win32":
            raise AttributeError
        fork = os.fork
    except AttributeError:
        sys.exit("Error: Must use Unix to be able to fork processes.")

    parser = ArgumentParser()
    add = parser.add_argument
    add(
        "-u",
        "--url",
        action="store",
        dest="url",
        help="start at the given URL before each script",
    )
    add(
        "-n",
        "--number",
        action="store",
        dest="number",
        default=1,
        type=int,
        help="number of times to run the given script(s)",
    )
    add(
        "-p",
        "--processes",
        action="store",
        dest="processes",
        default=1,
        type=int,
        help="number of processes to execute in parallel",
    )
    add(
        "scripts",
        metavar="SCRIPT",
        nargs="+",
        help="one or more twill scripts to execute",
    )

    args = parser.parse_args()

    # make sure that the current working directory is in the path
    if "" not in sys.path:
        sys.path.append("")

    average_number = args.number // args.processes
    last_number = average_number + args.number % args.processes
    child_pids = []
    is_parent = True
    repeat = 0

    # start a bunch of child processes and record their pids in the parent
    for i in range(args.processes):
        pid = fork()
        if pid:
            child_pids.append(pid)
        else:
            repeat = average_number if i else last_number
            is_parent = False
            break

    # set the children up to run and record their stats
    failed = False

    if is_parent:
        time.sleep(1)

        total_time = total_exec = 0

        # iterate over all the child pids, wait until they finish,
        # and then sum statistics
        for child_pid in child_pids[:]:
            pid, status = os.waitpid(child_pid, 0)
            if status or pid != child_pid:  # failure
                print(
                    f"[twill-fork parent: process {child_pid} FAILED:"
                    f" exit status {status}]"
                )
                print(
                    "[twill-fork parent:"
                    " (not counting stats for this process)]"
                )
                failed = True
            else:  # record statistics, otherwise
                filename = f".status.{child_pid}"
                with open(filename, "rb") as fp:
                    this_time, n_executed = load(fp)  # noqa: S301
                os.unlink(filename)  # noqa: PTH108
                total_time += this_time
                total_exec += n_executed

        # summarize
        print("\n----\n")
        print(f"number of processes: {args.processes}")
        print(f"total executed: {total_exec}")
        print(f"total time to execute: {total_time:.2f} s")
        if total_exec:
            avg_time = 1000 * total_time / total_exec
            print(f"average time: {avg_time:.2f} ms")
        else:
            print("(nothing completed, no average!)")
        print()

    else:
        pid = os.getpid()
        print(f"[twill-fork: pid {pid} : executing {repeat} times]")

        start_time = time.time()

        set_log_level("warning")
        for _i in range(repeat):
            for filename in args.scripts:
                execute_file(filename, initial_url=args.url)

        end_time = time.time()
        this_time = end_time - start_time

        # write statistics
        filename = f".status.{pid}"
        with open(filename, "wb") as fp:
            info = (this_time, repeat)
            dump(info, fp)

    sys.exit(-1 if failed else 0)


if __name__ == "__main__":
    main()



"""Global and local dictionaries, and initialization/utility functions."""

from typing import Any, Dict, Tuple

from . import errors

global_dict: Dict[str, Any] = {}  # the global dictionary


def init_global_dict() -> None:
    """Initialize the global dictionary with twill commands.

    This must be done after all the other modules are loaded, so that all
    the commands are already defined.
    """
    from . import commands, parse  # noqa: PLC0415

    cmd_list = commands.__all__
    global_dict.update((cmd, getattr(commands, cmd)) for cmd in cmd_list)
    parse.command_list.extend(cmd_list)


_local_dict_stack = []  # local dictionaries


def new_local_dict() -> Dict[str, Any]:
    """Initialize a new local dictionary and push it onto the stack."""
    local_dict: Dict[str, Any] = {}
    _local_dict_stack.append(local_dict)
    return local_dict


def pop_local_dict() -> Dict[str, Any]:
    """Get rid of the current local dictionary."""
    return _local_dict_stack.pop()


def get_twill_glocals() -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Return both global and current local dictionary."""
    if global_dict is None:
        raise errors.TwillException("Must initialize global namespace first!")
    if not _local_dict_stack:
        new_local_dict()
    return global_dict, _local_dict_stack[-1]



"""Map of various User-Agent string shortcuts that can be used for testing."""

from typing import Dict

agents: Dict[str, str] = {
    # Desktop
    "chrome_40": "Mozilla/5.0 (Windows NT 10.0) AppleWebKit/537.36"
    " (KHTML, like Gecko) Chrome/40.0.2214.93 Safari/537.36",
    "chrome_107": "Mozilla/5.0 (Windows NT 10.0; Win64; x64)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/107.0.0.0 Safari/537.36",
    "edge_12": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
    " (KHTML, like Gecko) Chrome/42.0.2311.135"
    " Safari/537.36 Edge/12.246",
    "edge_107": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36"
    " (KHTML, like Gecko) Chrome/107.0.0.0"
    " Safari/537.36 Edg/107.0.1418.26",
    "firefox_40": "Mozilla/5.0 (Windows NT 6.1; WOW64; rv:40.0)"
    " Gecko/20100101 Firefox/40.1",
    "firefox_106": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:106.0)"
    " Gecko/20100101 Firefox/106.0",
    "ie_3": "Mozilla/2.0 (compatible; MSIE 3.0; Windows 3.1)",
    "ie_4": "Mozilla/4.0 (compatible; MSIE 4.0; Windows NT 5.0)",
    "ie_5": "Mozilla/4.0 (compatible; MSIE 5.0; Windows NT 5.0)",
    "ie_6": "Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1)",
    "ie_7": "Mozilla/4.0 (compatible; MSIE 7.0; Windows NT 5.1)",
    "ie_8": "Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.0; Trident/4.0)",
    "ie_9": "Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0)",
    "ie_10": "Mozilla/5.0"
    " (compatible; MSIE 10.0; Windows NT 6.1; Trident/6.0)",
    "ie_11": "Mozilla/5.0 (Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko",
    "ie_mobile_9": "Mozilla/5.0 (compatible; MSIE 9.0; Windows Phone OS 7.5;"
    " Trident/5.0; IEMobile/9.0)",
    "opera_7": "Opera/7.0 (Windows NT 5.1; U) [en]",
    "opera_8": "Opera/8.00 (Windows NT 5.1; U; en)",
    "opera_9": "Opera/9.00 (Windows NT 5.2; U; en)",
    "opera_10": "Opera/9.80 (Windows NT 6.1; U; en)"
    " Presto/2.2.15 Version/10.00",
    "opera_11": "Opera/9.80 (Windows NT 6.1; U; en)"
    " Presto/2.7.62 Version/11.00",
    "opera_12": "Opera/12.0 (Windows NT 5.1; U; en)"
    " Presto/22.9.168 Version/12.00",
    "opera_mini_7": "Opera/9.80 (Android; Opera Mini/7.0.29952/28.2075; en)"
    " Presto/2.8.119 Version/11.10",
    "opera_mini_9": "Opera/9.80 (J2ME/MIDP; Opera Mini/9"
    " (Compatible; MSIE:9.0; iPhone; BlackBerry9700;"
    " AppleWebKit/24.746; en) Presto/2.5.25 Version/10.54",
    "konqueror_3": "Mozilla/5.0 (compatible; Konqueror/3.0; Linux)",
    "konqueror_4": "Mozilla/5.0 (compatible; Konqueror/4.0; Linux)"
    " KHTML/4.0.3 (like Gecko)",
    "lynx_2_8": "Lynx/2.8.7rel.2 libwww-FM/2.14 SSL-MM/1.4.1 OpenSSL/1.0.0a",
    "w3m_0_5": "w3m/0.5.2 (Linux i686; it; Debian-3.0.6-3)",
    "netscape_3": "Mozilla/3.0 (X11; I; AIX 2)",
    "netscape_4": "Mozilla/4.0 (compatible; Mozilla/5.0 ; Linux i686)",
    "netscape_4_5": "Mozilla/4.5 [en] (X11; I; SunOS 5.6 sun4u)",
    "netscape_7": "Mozilla/5.0 (X11; U; SunOS sun4u; en-US; rv:1.0.1)"
    " Gecko/20020921 Netscape/7.0",
    "netscape_9": "Mozilla/5.0 (X11; U; Linux i686; en-US; rv:1.8.1.8pre)"
    " Gecko/20071015 Firefox/2.0.0.7 Navigator/9.0",
    "palemoon_25": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:25.6)"
    " Gecko/20150723 Firefox/31.9 PaleMoon/25.6.0",
    "safari_1": "Mozilla/5.0 (Macintosh; PPC Mac OS X; en)"
    " AppleWebKit/85.7 (KHTML, like Gecko) Safari/85.6",
    "safari_2": "Mozilla/5.0 (Macintosh; PPC Mac OS; en)"
    " AppleWebKit/412 (KHTML, like Gecko) Safari/412",
    "safari_3": "Mozilla/5.0 (Macintosh; Intel Mac OS X; en)"
    " AppleWebKit/522.7 (KHTML, like Gecko)"
    " Version/3.0 Safari/522.7",
    "safari_4": "Mozilla/5.0 (Macintosh; U; PPC Mac OS X 10_5_6; en)"
    " AppleWebKit/530.9+ (KHTML, like Gecko)"
    "Version/4.0 Safari/528.16",
    "safari_5": "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10_6_3; en)"
    " AppleWebKit/534.1+ (KHTML, like Gecko)"
    " Version/5.0 Safari/533.16",
    "safari_6": "Mozilla/5.0 (iPad; CPU OS 6_0 like Mac OS X)"
    " AppleWebKit/536.26 (KHTML, like Gecko)"
    " Version/6.0 Mobile/10A5355d Safari/8536.25",
    "safari_7": "Mozilla/5.0 (iPad; CPU OS 7_1_2 like Mac OS X)"
    " AppleWebKit/537.51.2 (KHTML, like Gecko)"
    " Version/7.0 Mobile/11D257 Safari/9537.53",
    "safari_605": "Mozilla/5.0 (Macintosh; Intel Mac OS X 13_0)"
    " AppleWebKit/605.1.15 (KHTML, like Gecko)"
    " Version/16.1 Safari/605.1.15",
    "vivaldi_5": "Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36"
    " (KHTML, like Gecko) Chrome/107.0.0.0"
    " Safari/537.36 Vivaldi/5.4.2753.51",
    # Android phones
    "galaxy_s7": "Mozilla/5.0 (Linux; Android 7.0; SM-G930VC Build/NRD90M; wv)"
    " AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0"
    " Chrome/58.0.3029.83 Mobile Safari/537.36",
    "galaxy_s10": "Mozilla/5.0"
    " (Linux; Android 9; SM-G973U Build/PPR1.180610.011)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/69.0.3497.100 Mobile Safari/537.36",
    "galaxy_s20": "Mozilla/5.0 (Linux; Android 10;"
    " SM-G980F Build/QP1A.190711.020; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.96"
    " Mobile Safari/537.36",
    "galaxy_s22": "Mozilla/5.0 (Linux; Android 12;"
    " SM-S906N Build/QP1A.190711.020; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119"
    " Mobile Safari/537.36",
    "google_pixel": "Mozilla/5.0 (Linux; Android 7.1.1; Google Pixel"
    " Build/NMF26F; wv) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Version/4.0 Chrome/54.0.2840.85 Mobile Safari/537.36",
    "google_pixel4": "Mozilla/5.0 (Linux; Android 10; Google Pixel 4"
    " Build/QD1A.190821.014.C2; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0 Chrome/78.0.3904.108"
    " Mobile Safari/537.36",
    "google_pixel_6": "Mozilla/5.0 (Linux; Android 12; Pixel 6"
    " Build/SD1A.210817.023; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0 Chrome/94.0.4606.71"
    " Mobile Safari/537.36",
    "nexus_6p": "Mozilla/5.0 (Linux; Android 6.0.1; Nexus 6P Build/MMB29P)"
    " AppleWebKit/537.36 (KHTML, like Gecko) Chrome/47.0.2526.83"
    " Mobile Safari/537.36",
    "sony_xperia_1": "Mozilla/5.0 (Linux; Android 9;"
    " J8110 Build/55.0.A.0.552; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0 Chrome/71.0.3578.99"
    " Mobile Safari/537.36",
    "htc_one_x10": "Mozilla/5.0 (Linux; Android 6.0; HTC One"
    " X10 Build/MRA58K; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0 Chrome/61.0.3163.98"
    " Mobile Safari/537.36",
    # iPhones
    "iphone_6": "Mozilla/5.0"
    " (Apple-iPhone7C2/1202.466; U; CPU like Mac OS X; en)"
    " AppleWebKit/420+ (KHTML, like Gecko) Version/3.0"
    " Mobile/1A543 Safari/419.3",
    "iphone_7": "Mozilla/5.0"
    " (iPhone9,3; U; CPU iPhone OS 10_0_1 like Mac OS X)"
    " AppleWebKit/602.1.50 (KHTML, like Gecko) Version/10.0"
    " Mobile/14A403 Safari/602.1",
    "iphone_8": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X)"
    " AppleWebKit/604.1.34 (KHTML, like Gecko) Version/11.0"
    " Mobile/15A5341f Safari/604.1",
    "iphone_x": "Mozilla/5.0 (iPhone; CPU iPhone OS 11_0 like Mac OS X)"
    " AppleWebKit/604.1.38 (KHTML, like Gecko) Version/11.0"
    " Mobile/15A372 Safari/604.1",
    "iphone_11": "Mozilla/5.0"
    " (iPhone12,1; U; CPU iPhone OS 13_0 like Mac OS X)"
    " AppleWebKit/602.1.50 (KHTML, like Gecko)"
    " Version/10.0 Mobile/15E148 Safari/602.1",
    "iphone_12": "Mozilla/5.0"
    " (iPhone13,2; U; CPU iPhone OS 14_0 like Mac OS X)"
    " AppleWebKit/602.1.50 (KHTML, like Gecko)"
    " Version/10.0 Mobile/15E148 Safari/602.1",
    "iphone_13_pro_max": "Mozilla/5.0 (iPhone14,3; U; CPU iPhone OS 15_0"
    " like Mac OS X) AppleWebKit/602.1.50"
    " (KHTML, like Gecko) Version/10.0"
    " Mobile/19A346 Safari/602.1",
    "iphone_se_3": "Mozilla/5.0 (iPhone14,6; U; CPU iPhone OS 15_4"
    " like Mac OS X) AppleWebKit/602.1.50 (KHTML, like Gecko)"
    " Version/10.0 Mobile/19E241 Safari/602.1",
    # MS Windows phones
    "ms_lumia_650": "Mozilla/5.0"
    " (Windows Phone 10.0; Android 6.0.1; Microsoft; RM-1152)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/52.0.2743.116 Mobile Safari/537.36 Edge/15.15254",
    "ms_lumia_950": "Mozilla/5.0 (Windows Phone 10.0;"
    " Android 4.2.1; Microsoft; Lumia 950)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/46.0.2486.0 Mobile Safari/537.36 Edge/13.1058",
    # Tablets
    "galaxy_tab_s8": "Mozilla/5.0 (Linux; Android 12;"
    " SM-X906C Build/QP1A.190711.020; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0 Chrome/80.0.3987.119"
    " Mobile Safari/537.36",
    "lenovo_yoga_tab_11": "Mozilla/5.0 (Linux; Android 11; Lenovo YT-J706X)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/96.0.4664.45 Safari/537.36",
    "sony_xperia_tab_z4": "Mozilla/5.0 (Linux; Android 6.0.1;"
    " SGP771 Build/32.2.A.0.253; wv) AppleWebKit/537.36"
    " (KHTML, like Gecko) Version/4.0"
    " Chrome/52.0.2743.98 Safari/537.36",
    "galaxy_tab_s3": "Mozilla/5.0 (Linux; Android 7.0; SM-T827R4 Build/NRD90M)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/60.0.3112.116 Safari/537.36",
    "amazon_fire_hdx_7": "Mozilla/5.0"
    " (Linux; Android 4.4.3; KFTHWI Build/KTU84M)"
    " AppleWebKit/537.36 (KHTML, like Gecko) Silk/47.1.79"
    " like Chrome/47.0.2526.80 Safari/537.36",
    "lg_g_pad_7": "Mozilla/5.0 (Linux; Android 5.0.2; LG-V410/V41020c"
    " Build/LRX22G) AppleWebKit/537.36 (KHTML, like Gecko)"
    " Version/4.0 Chrome/34.0.1847.118 Safari/537.36",
    # E-Readers
    "kindle_4": "Mozilla/5.0 (X11; U; Linux armv7l like Android; en-us)"
    " AppleWebKit/531.2+ (KHTML, like Gecko) Version/5.0"
    " Safari/533.2+ Kindle/3.0+",
    "kindle_3": "Mozilla/5.0 (Linux; U; en-US) AppleWebKit/528.5+"
    " (KHTML, like Gecko, Safari/528.5+) Version/4.0 Kindle/3.0"
    " (screen 600x800; rotate)",
    # Set tops
    "chromecast": "Mozilla/5.0 (CrKey armv7l 1.5.16041) AppleWebKit/537.36"
    " (KHTML, like Gecko) Chrome/31.0.1650.0 Safari/537.36",
    "amazon_4k_fire_tv": "Mozilla/5.0 (Linux; Android 5.1; AFTS Build/LMY47O)"
    " AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0"
    " Chrome/41.99900.2250.0242 Safari/537.36",
    "nexus_player": "Dalvik/2.1.0 (Linux; U; Android 6.0.1;"
    " Nexus Player Build/MMB29T)",
    "apple_tv_6": "AppleTV11,1/11.1",
    "apple_tv_5": "AppleTV6,2/11.1",
    "apple_tv_4": "AppleTV5,3/9.1.1",
    # Game consoles
    "playstation_5": "Mozilla/5.0 (PlayStation; PlayStation 5/2.26)"
    " AppleWebKit/605.1.15 (KHTML, like Gecko)"
    " Version/13.0 Safari/605.1.15",
    "playstation_4": "Mozilla/5.0 (PlayStation 4 3.11) AppleWebKit/537.73"
    " (KHTML, like Gecko)",
    "xbox_x": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; Xbox; Xbox Series X)"
    " AppleWebKit/537.36 (KHTML, like Gecko)"
    " Chrome/48.0.2564.82 Safari/537.36 Edge/20.02",
    "xbox_one": "Mozilla/5.0"
    " (Windows Phone 10.0; Android 4.2.1; Xbox; Xbox One)"
    " AppleWebKit/537.36 (KHTML, like Gecko) Chrome/46.0.2486.0"
    " Mobile Safari/537.36 Edge/13.10586",
    "nintendo_switch": "Mozilla/5.0 (Nintendo Switch; WifiWebAuthApplet)"
    " AppleWebKit/601.6 (KHTML, like Gecko) NF/4.0.0.5.10"
    " NintendoBrowser/5.1.0.13343",
    # Bots
    "google_bot_2": "Mozilla/5.0 (compatible; Googlebot/2.1;"
    " +http://www.google.com/bot.html)",
    "bing_bot_2": "Mozilla/5.0 (compatible; bingbot/2.0;"
    " +http://www.bing.com/bingbot.htm)",
    "yahoo_bot": "Mozilla/5.0 (compatible; Yahoo! Slurp;"
    " http://help.yahoo.com/help/us/ysearch/slurp)",
}



"""Suresh's extension for slicing and dicing using regular expressions."""

import re
from typing import Any

from twill import browser, log
from twill.namespaces import get_twill_glocals


def showvar(which: str) -> None:
    """>> showvar var

    Shows the value of the variable 'var'.
    """
    global_dict, local_dict = get_twill_glocals()

    d = global_dict.copy()
    d.update(local_dict)

    log.info(d.get(str(which)))


def split(what: str) -> None:
    """>> split <regex>

    Sets __matchlist__ to re.split(regex, page).
    """
    page = browser.html

    m = re.split(what, page)

    global_dict, local_dict = get_twill_glocals()
    local_dict["__matchlist__"] = m


def findall(what: str) -> None:
    """>> findall <regex>

    Sets __matchlist__ to re.findall(regex, page).
    """
    page = browser.html

    regex = re.compile(what, re.DOTALL)
    m = regex.findall(page)

    global_dict, local_dict = get_twill_glocals()
    local_dict["__matchlist__"] = m


def getmatch(where: str, what: str) -> None:
    """>> getmatch into_var expression

    Evaluates an expression against __match__ and puts it into 'into_var'.
    """
    global_dict, local_dict = get_twill_glocals()
    match = local_dict["__match__"]
    local_dict[where] = _eval(match, what)


def setmatch(what: str) -> None:
    """>> setmatch expression

    Sets each element __matchlist__ to eval(expression); 'm' is set
    to each element of __matchlist__ prior to processing.
    """
    global_dict, local_dict = get_twill_glocals()

    match = local_dict["__matchlist__"]
    if isinstance(match, str):
        match = [match]

    new_match = [_eval(m, what) for m in match]
    local_dict["__matchlist__"] = new_match


def _eval(match: str, exp: str) -> Any:
    """Evaluate an expression."""
    return eval(exp, globals(), {"m": match})  # noqa: S307


def popmatch(which: str) -> None:
    """>> popmatch index

    Pops __matchlist__[i] into __match__.
    """
    global_dict, local_dict = get_twill_glocals()

    matchlist = local_dict["__matchlist__"]
    match = matchlist.pop(int(which))
    local_dict["__match__"] = match



"""The twill shell commands.

Implementation of all the individual twill commands that are available
through the twill command line script.
"""

import getpass
import re
import sys
import time
from os.path import sep
from typing import Any, Dict, Optional

from . import log, set_err_out, set_output, utils
from .agents import agents
from .browser import browser
from .errors import TwillAssertionError, TwillException
from .namespaces import get_twill_glocals

__all__ = [
    "add_auth",
    "add_cleanup",
    "add_extra_header",
    "agent",
    "back",
    "browser",
    "clear_cookies",
    "clear_extra_headers",
    "code",
    "config",
    "debug",
    "echo",
    "exit",
    "extend_with",
    "fa",
    "find",
    "follow",
    "form_action",
    "form_clear",
    "form_file",
    "form_value",
    "formaction",
    "formclear",
    "formfile",
    "formvalue",
    "fv",
    "get_input",
    "get_password",
    "getinput",
    "getpassword",
    "go",
    "info",
    "load_cookies",
    "not_find",
    "notfind",
    "options",
    "redirect_error",
    "redirect_output",
    "reload",
    "reset_browser",
    "reset_error",
    "reset_output",
    "rf",
    "run",
    "run_file",
    "runfile",
    "save_cookies",
    "save_html",
    "set_global",
    "set_local",
    "setglobal",
    "setlocal",
    "show",
    "show_cookies",
    "show_extra_headers",
    "show_forms",
    "show_history",
    "show_html",
    "show_links",
    "showcookies",
    "showforms",
    "showhistory",
    "showhtml",
    "showlinks",
    "sleep",
    "submit",
    "tidy_ok",
    "timeout",
    "title",
    "url",
]


def reset_browser(base_url: str = "") -> None:
    """>> reset_browser [base_url]

    Reset the browser completely.
    """
    browser.reset(base_url=base_url)
    options.clear()
    options.update(default_options)


def exit(code: str = "0") -> None:
    """>> exit [<code>]

    Exit twill with given exit code (default 0, "no error").
    """
    raise SystemExit(int(code) if code else 0)


def go(url: str) -> None:
    """>> go <url>

    Visit the URL given.
    """
    browser.go(url)


def reload() -> None:
    """>> reload

    Reload the current URL.
    """
    browser.reload()


def timeout(timeout: float = 10) -> None:
    """>> timeout [timeout]

    Set the request timeout (0 for no timeout).
    """
    if isinstance(timeout, str):
        timeout = float(timeout) if timeout else 0
    browser.timeout = timeout or None


def code(should_be: int) -> None:
    """>> code <int>

    Check to make sure the response code for the last page is as given.
    """
    if not isinstance(should_be, int):
        should_be = int(should_be)
    if browser.code != should_be:
        raise TwillAssertionError(f"code is {browser.code} != {should_be}")


def tidy_ok() -> None:
    """>> tidy_ok

    Assert that 'tidy' does not produce any warnings or errors when run on
    the current page.

    If 'tidy' cannot be run, will fail silently (unless 'require_tidy' option
    is true; see 'config' command).
    """
    page = browser.html
    if page is None:
        raise TwillAssertionError("not viewing HTML!")

    clean_page, errors = utils.run_tidy(page)
    if clean_page is None:  # tidy doesn't exist...
        if options.get("require_tidy"):
            raise TwillAssertionError("cannot run 'tidy'")
    elif errors:
        raise TwillAssertionError(f"tidy errors:\n====\n{errors}\n====\n")


def url(should_be: str) -> str:
    """>> url <pattern>

    Check to make sure that the current URL matches the regex pattern.
    The local variable __match__ is set to the matching part of the URL.
    """
    regex = re.compile(should_be)
    current_url = browser.url

    if current_url is None:
        current_url = ""
        m = None
    else:
        m = regex.search(current_url)

    if not m:
        raise TwillAssertionError(
            f"current url is '{current_url}';\ndoes not match '{should_be}'\n"
        )

    match_str = m.group(1 if m.groups() else 0)
    global_dict, local_dict = get_twill_glocals()
    local_dict["__match__"] = match_str
    return match_str


def follow(what: str) -> str:
    """>> follow <pattern>

    Find the first link on the page matching the given regex pattern and
    then visit it.
    """
    link = browser.find_link(what)
    if link:
        browser.follow_link(link)
        if not browser.url:
            raise TwillAssertionError(f"Cannot follow link '{link}'")
        return browser.url

    raise TwillAssertionError(f"no links match to '{what}'")


_find_flags = {"i": re.IGNORECASE, "m": re.MULTILINE, "s": re.DOTALL}


def _parse_find_flags(flags: str) -> int:
    """Helper function to parse the find flags."""
    re_flags = 0
    for char in flags:
        try:
            re_flags |= _find_flags[char]
        except IndexError as error:
            raise TwillAssertionError(
                f"unknown 'find' flag {char!r}"
            ) from error
    return re_flags


def find(what: str, flags: str = "") -> str:
    """>> find <pattern> [<flags>]

    Succeed if the regular expression pattern can be found on the page.
    Sets the local variable __match__ to the matching text.

    Flags is a string consisting of the following characters:

    * i: ignore case
    * m: multi-line
    * s: dot matches all
    * x: use XPath expressions instead of regular expression

    For explanations of regular expressions, please see the Python re module
    documentation.
    """
    page = browser.html
    local_dict = get_twill_glocals()[1]
    if "x" in flags:
        elements = browser.xpath(what)
        if not elements:
            raise TwillAssertionError(f"no element to path '{what}'")
        match_str = elements[0].text or ""
    else:
        match = re.search(what, page, flags=_parse_find_flags(flags))
        if not match:
            raise TwillAssertionError(f"no match to '{what}'")
        match_str = match.group(1 if match.groups() else 0)
    local_dict["__match__"] = match_str
    return match_str


def not_find(what: str, flags: str = "") -> None:
    """>> not_find <pattern> [<flags>]

    Fail if the regular expression pattern can be found on the page.
    """
    try:
        find(what, flags)
    except TwillAssertionError:
        pass
    else:
        raise TwillAssertionError(f"match to '{what}'")


notfind = not_find  # backward compatibility and convenience


def back() -> None:
    """>> back

    Return to the previous page.
    """
    browser.back()


def show(what: Optional[str] = None) -> None:
    """>> show [<objects>]

    Show the specified objects (html, cookies, forms, links, history).
    """
    if not what:
        what = "html"
    command = None
    if what.isalpha():
        command_name = f"show_{what}"
        if command_name in __all__:
            command = globals().get(command_name)
    if not command:
        raise TwillException(f'Cannot show "{what}".')
    command()


def show_html() -> None:
    """>> show_html

    Show the HTML for the current page or show the specified objects
    (which can be cookies, forms, history or links).

    Note: Use browser.html to get the HTML programmatically.
    """
    html = browser.html.strip()
    log.info("")
    log.info(html)
    log.info("")


showhtml = show_html  # backward compatibility and consistency


def echo(*strs: str) -> None:
    """>> echo <list> <of> <strings>

    Echo the arguments to the screen.
    """
    log.info(" ".join(map(str, strs)))


def save_html(filename: Optional[str] = None) -> None:
    """>> save_html [<filename>]

    Save the HTML for the current page into <filename>.
    If no filename given, construct the filename from the URL.
    """
    html = browser.html
    if html is None:
        log.warning("No page to save.")
        return

    if filename is None:
        url = browser.url
        if url:
            url = url.split("?", 1)[0]
            filename = url.rsplit("/", 1)[-1]
        if not filename:
            filename = "index.html"
        log.info("Using filename '%s'.", filename)

    encoding = browser.encoding or "utf-8"
    try:
        with open(filename, "w", encoding=encoding) as f:
            f.write(html)
    except UnicodeEncodeError:
        if encoding == "utf-8":
            raise
        with open(filename, "w", encoding="utf-8") as f:
            f.write(html)


def sleep(interval: str = "1") -> None:
    """>> sleep [<interval>]

    Sleep for the specified amount of time.
    If no interval is given, sleep for 1 second.
    """
    time.sleep(float(interval))


def agent(what: str) -> None:
    """>> agent <agent>

    Set the agent string (identifying the browser brand).

    Some convenient shortcuts:
    chrome_107, firefox_106, safari_605, edge_107, ie_11.

    See twill.agents for a list of all available shortcuts.
    """
    what = what.strip()
    agent = agents.get(what, what)
    browser.agent_string = agent


def submit(
    submit_button: Optional[str] = None, form_name: Optional[str] = None
) -> None:
    """>> submit [<submit_button> [<form_name>]]

    Submit the current form (the one last clicked on) by clicking on the
    given submission button.  If no 'submit_button' is given, submit the
    current form by using the last clicked submit button.

    The form to submit is the last form clicked on with a 'form_value' command
    unless explicitly specified given the

    The button used to submit is chosen based on 'submit_button'.
    If 'submit_button' is given, it's matched against buttons using
    the same rules that 'form_value' uses.  If 'button_name' is not given,
    this function uses the last submit button clicked on by 'form_value'.
    If none can be found, it submits the form with no submit button clicked.
    """
    browser.submit(submit_button, form_name)


def show_forms() -> None:
    """>> show_forms

    Show all the forms on the current page.

    Note: Use browser.forms to get the forms programmatically.
    """
    browser.show_forms()


showforms = show_forms  # backward compatibility and convenience


def show_links() -> None:
    """>> show_links

    Show all the links on the current page.

    Note: Use browser.links to get the links programmatically.
    """
    browser.show_links()


showlinks = show_links  # backward compatibility and convenience


def show_history() -> None:
    """>> show_history

    Show the browser history (what URLs were visited).

    Note: Use browser.history to get the history programmatically.
    """
    browser.show_history()


showhistory = show_history  # backward compatibility and convenience


def form_clear(form_name: str) -> None:
    """>> form_clear <form_name>

    Run 'clear' on all the controls in this form.
    """
    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")
    for control in form.inputs:
        if not (
            "readonly" in control.attrib
            or "disabled" in control.attrib
            or getattr(control, "type", None) in ("submit", "image", "hidden")
        ):
            del control.value
    browser.last_submit_button = None


formclear = form_clear  # backward compatibility and convenience


def form_value(form_name: str, field_name: str, value: str) -> None:
    """>> form_value <form_name> <field_name> <value>

    Set value of a form field.

    There are some ambiguities in the way 'form_value' deals with lists:
    'form_value' will *add* the given value to a list of multiple selection,
    for lists that allow it.

    Forms are matched against 'form_name' as follows:
      1. regex match to actual form name;
      2. if 'form_name' is an integer, it's tried as an index.

    Form controls are matched against 'field_name' as follows:
      1. unique exact match to control name;
      2. unique regex match to control name;
      3. if field_name is an integer, it's tried as an index;
      4. unique & exact match to submit-button values.

    'form_value' ignores read-only fields completely; if they're readonly,
    nothing is done, unless the config options ('config' command) are
    changed.

    'form_value' is available as 'fv' as well.
    """
    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")

    control = browser.form_field(form, field_name)

    browser.clicked(form, control)

    attrib = getattr(control, "attrib", {})
    if "readonly" in attrib:
        if options["readonly_controls_writeable"]:
            log.info("Forcing read-only form field to writeable.")
            del attrib["readonly"]
        else:
            log.info("Form field is read-only or ignorable; nothing done.")
            return

    if getattr(control, "type", None) == "file":
        raise TwillException(
            'form field is for file upload; use "form_file" instead'
        )

    value = browser.decode(value)
    utils.set_form_control_value(control, value)


fv = formvalue = form_value  # backward compatibility and convenience


def form_action(form_name: str, action_url: str) -> None:
    """>> form_action <form_name> <action_url>

    Sets action parameter on form to action_url.

    'form_action' is available as 'fa' as well.
    """
    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")
    log.info("Setting action for form %s to %s.", form, action_url)
    form.action = action_url


fa = formaction = form_action  # backward compatibility and convenience


def form_file(
    form_name: str,
    field_name: str,
    filename: str,
    content_type: Optional[str] = None,
) -> None:
    """>> form_file <form_name> <field_name> <filename> [<content_type>]

    Upload a file via an "upload file" form field.
    """
    filename = filename.replace("/", sep)

    form = browser.form(form_name)
    if form is None:
        raise TwillAssertionError("Form not found")
    control = browser.form_field(form, field_name)

    if getattr(control, "type", None) != "file":
        raise TwillException("ERROR: field is not a file upload field!")

    browser.clicked(form, control)
    plain = content_type and content_type.startswith(("plain/", "html/"))
    fp = open(filename, "r" if plain else "rb")  # noqa: SIM115
    browser.add_form_file(field_name, fp)

    log.info(
        'Added file "%s" to file upload field "%s".', filename, field_name
    )


formfile = form_file  # backward compatibility and convenience


def extend_with(module_name: str) -> None:
    """>> extend_with <module_name>

    Import contents of given module.
    """
    global_dict, local_dict = get_twill_glocals()

    exec(f"from {module_name} import *", global_dict)

    # now add the commands into the commands available for the shell,
    # and print out some nice stuff about what the extension module does.

    mod = sys.modules[module_name]

    from . import parse, shell  # noqa: PLC0415

    fn_list = getattr(mod, "__all__", None)
    if fn_list is None:
        fn_list = [fn for fn in dir(mod) if callable(getattr(mod, fn))]

    for command in fn_list:
        fn = getattr(mod, command)
        shell.add_command(command, fn.__doc__)
        parse.command_list.append(command)

    info, debug = log.info, log.debug
    info("Imported extension module '%s'.", module_name)
    debug("(at %s)", mod.__file__)

    if shell.interactive:
        if mod.__doc__:
            info("\nDescription:\n\n%s\n", mod.__doc__.strip())
        elif fn_list:
            info("New commands:\n")
            for name in fn_list:
                info("\t%s", name)
            info("")


def get_input(prompt: str) -> str:
    """>> get_input <prompt>

    Get input, store it in '__input__'.
    """
    local_dict = get_twill_glocals()[1]

    inp = input(prompt)

    local_dict["__input__"] = inp
    return inp


getinput = get_input  # backward compatibility and convenience


def get_password(prompt: str) -> str:
    """>> get_password <prompt>

    Get a password ("invisible input"), store it in '__password__'.
    """
    local_dict = get_twill_glocals()[1]

    # we use sys.stdin here in order to get the same behaviour on Unix
    # as on other platforms and for better testability of this function
    inp = getpass.getpass(prompt, sys.stdin)

    local_dict["__password__"] = inp
    return inp


getpassword = get_password  # backward compatibility and convenience


def save_cookies(filename: str) -> None:
    """>> save_cookies <filename>

    Save all the current cookies to the given file.
    """
    browser.save_cookies(filename)


def load_cookies(filename: str) -> None:
    """>> load_cookies <filename>

    Clear the cookie jar and load cookies from the given file.
    """
    browser.load_cookies(filename)


def clear_cookies() -> None:
    """>> clear_cookies

    Clear the cookie jar.
    """
    browser.clear_cookies()


def show_cookies() -> None:
    """>> show_cookies

    Show all the cookies in the cookie jar.

    Note: Use browser.cookies to get the cookies programmatically.
    """
    browser.show_cookies()


showcookies = show_cookies  # backward compatibility and convenience


def add_auth(realm: str, uri: str, user: str, passwd: str) -> None:
    """>> add_auth <realm> <uri> <user> <passwd>

    Add HTTP Basic Authentication information for the given realm/uri.
    """
    if realm is not None:
        browser.add_creds((uri, realm), user, passwd)
        log.info(
            "Added auth info: realm '%s' / URI '%s' / user '%s'.",
            realm,
            uri,
            user,
        )
    if realm is None or options["with_default_realm"]:
        browser.add_creds(uri, user, passwd)
        if realm is None:
            log.info("Added auth info: URI '%s' / user '%s'.", uri, user)


def debug(what: str, level: str) -> None:
    """>> debug <what> <level>

    <what> can be:
       * http (any level >= 1), to display the HTTP transactions.
       * commands (any level >= 1), to display the commands being executed.
       * equiv-refresh (any level >= 1) to display HTTP-EQUIV refresh handling.
    """
    from . import parse  # noqa: PLC0415

    try:
        num_level = int(level)
    except ValueError:
        num_level = 1 if utils.make_boolean(level) else 0

    log.info("DEBUG: Setting %s debugging to level %d.", what, num_level)

    if what == "http":
        browser.debug_level = num_level
    elif what == "equiv-refresh":
        browser.show_refresh = num_level > 0
    elif what == "commands":
        parse.log_commands(num_level > 0)
    else:
        raise TwillException(f'Unknown debugging type: "{what}"')


def run(cmd: str) -> None:
    """>> run <command>

    <command> can be any valid Python command; 'exec' is used to run it.
    """
    # @CTB: use pyparsing to grok the command?  make sure that quoting works...

    # execute command.
    global_dict, local_dict = get_twill_glocals()

    # set __url__
    local_dict["__cmd__"] = cmd
    local_dict["__url__"] = browser.url

    exec(cmd, global_dict, local_dict)


def run_file(*args: str) -> None:
    """>> run_file <file1> [<file2> ...]

    Execute the given twill scripts or directories of twill scripts.

    'run_file' is available as 'rf' as well.
    """
    from . import parse  # noqa: PLC0415

    filenames = utils.gather_filenames(args)
    for filename in filenames:
        parse.execute_file(filename, no_reset=True)


rf = runfile = run_file  # backward compatibility and convenience


def add_cleanup(*args: str) -> None:
    """>> add_cleanup <file1> [<file2> ...]

    Execute the given twill scripts after the current twill script.
    """
    local_dict = get_twill_glocals()[1]
    cleanups = local_dict.setdefault("__cleanups__", [])
    filenames = utils.gather_filenames(args)
    log.debug("Adding cleanup scripts: %s", ", ".join(filenames))
    cleanups.extend(reversed(filenames))


def set_global(name: str, value: str) -> None:
    """set_global <name> <value>

    Sets the variable <name> to the value <value> in the global namespace.
    """
    global_dict, local_dict = get_twill_glocals()
    global_dict[name] = value


setglobal = set_global  # backward compatibility and convenience


def set_local(name: str, value: str) -> None:
    """set_local <name> <value>

    Sets the variable <name> to the value <value> in the local namespace.
    """
    global_dict, local_dict = get_twill_glocals()
    local_dict[name] = value


setlocal = set_local  # backward compatibility and convenience


def title(what: str) -> str:
    """>> title <pattern>

    Succeed if the regular expression pattern is in the page title.
    """
    regex = re.compile(what)
    title = browser.title

    if title is None:
        log.info("The page has no title.")
    else:
        log.info("The title is '%s'.", title)

    m = regex.search(title) if title else None
    if m is None:
        raise TwillAssertionError(f"The title does not contain '{what}'.")

    match_str = m.group(1 if m.groups() else 0)

    global_dict, local_dict = get_twill_glocals()
    local_dict["__match__"] = match_str
    return match_str


def redirect_output(filename: str) -> None:
    """>> redirect_output <filename>

    Append all twill output to the given file.
    """
    fp = open(filename, "a", encoding="utf-8")  # noqa: SIM115
    set_output(fp)


def reset_output() -> None:
    """>> reset_output

    Reset twill output to go to the screen.
    """
    set_output(None)


def redirect_error(filename: str) -> None:
    """>> redirect_error <filename>

    Append all twill error output to the given file.
    """
    fp = open(filename, "a", encoding="utf-8")  # noqa: SIM115
    set_err_out(fp)


def reset_error() -> None:
    """>> reset_error

    Reset twill error output to go to the screen.
    """
    set_err_out(None)


def add_extra_header(header_key: str, header_value: str) -> None:
    """>> add_header <name> <value>

    Add an HTTP header to each HTTP request.  See 'show_extra_headers' and
    'clear_extra_headers'.
    """
    browser.headers[header_key] = header_value


def show_extra_headers() -> None:
    """>> show_extra_headers

    Show any extra headers being added to each HTTP request.
    """
    info = log.info
    headers = browser.headers
    if headers:
        info("\nThe following HTTP headers are added to each request:\n")
        for key, value in headers.items():
            info('\t"%s" = "%s"', key, value)
        info("")
    else:
        info("** no extra HTTP headers **")


def clear_extra_headers() -> None:
    """>> clear_extra_headers

    Remove all user-defined HTTP headers.  See 'add_extra_header' and
    'show_extra_headers'.
    """
    browser.reset_headers()


default_options: Dict[str, Any] = {
    "equiv_refresh_interval": 2,
    "readonly_controls_writeable": False,
    "require_tidy": False,
    "with_default_realm": False,
}

options = default_options.copy()  # the global options dictionary


def config(key: Optional[str] = None, value: Any = None) -> None:
    """>> config [<key> [<int value>]]

    Configure/report various options.  If no <value> is given, report
    the current key value; if no <key> given, report current settings.

    Options starting with "tidy_" will be used to configure HTML tidy.

    So far:

     * 'equiv_refresh_interval', default 2 -- time limit for HTTP-EQUIV=REFRESH
     * 'readonly_controls_writeable', default False -- all controls writeable
     * 'require_tidy', default False -- *require* that tidy be installed
     * 'with_default_realm', default False -- use a default realm for HTTP AUTH
    """
    info = log.info
    if key is None:
        keys = sorted(options)
        info("\nCurrent configuration:\n")
        for k in keys:
            info("\t%s : %s", k, options[k])
        info("")
    else:
        v = options.get(key)
        if v is None and not key.startswith("tidy_"):
            log.error("no such configuration key '%s'", key)
            info("valid keys are: %s", sorted(options))
            raise TwillException(f"no such configuration key: {key!r}")
        if value is None:
            info("\nkey %s: value %s\n", key, v)
        else:
            if isinstance(v, bool):
                value = utils.make_boolean(value)
            elif isinstance(v, int):
                value = utils.make_int(value)
            options[key] = value


def info() -> None:
    """>> info

    Report information on current page.
    """
    current_url = browser.url
    if current_url is None:
        log.warning("We're not on a page!")
        return

    content_type = browser.response_headers["content-type"]
    if isinstance(content_type, bytes):
        content_type = content_type.decode()
    is_html = content_type and content_type.split(";", 1)[0] == "text/html"
    code = browser.code

    info = log.info
    info("\tURL: %s", current_url)
    info("\tHTTP code: %s", code)
    info("\tContent type: %s%s", content_type, " (HTML)" if is_html else "")
    if is_html:
        title = browser.title
        info("\tPage title: %s", title)
        forms = browser.forms
        if len(forms):
            info("\tThis page contains %d form(s)", len(forms))
    info("")



# This file is part of the twill source distribution.
#
# twill is an extensible scriptlet language for testing Web apps,
# available at https://github.com/twill-tools/twill.
#
# Copyright (c) 2005-2025
# by C. Titus Brown, Ben R. Taylor, Christoph Zwerschke et al.
#
# This program and all associated source code files are released under the
# terms of the MIT license; please see the included LICENSE file for more
# information, or go to https://opensource.org/licenses/mit-license.php.

"""The twill web browsing and testing language and associated utilities.

A scripting system for automating web browsing.  Useful for testing
web pages or grabbing data from password-protected sites automatically.
"""

import importlib.metadata
import logging
import sys
from pathlib import Path
from typing import Optional, TextIO, Union

metadata = importlib.metadata.metadata(__package__)

__version__: str = metadata["Version"]

__url__: str = metadata["Project-URL"].rsplit(None, 1)[-1]

__all__ = [
    "TwillCommandLoop",
    "__url__",
    "__version__",
    "browser",
    "execute_file",
    "execute_string",
    "log",
    "set_err_out",
    "set_log_level",
    "set_output",
    "twill_ext",
]

this_dir = Path(__file__).parent
# Add extensions directory at the *end* of sys.path.
# This means that user extensions will take priority over twill extensions.
extensions = this_dir / "extensions"
sys.path.append(str(extensions))

twill_ext = ".twill"  # file extension for twill scripts


log_levels = {
    "CRITICAL": logging.CRITICAL,
    "ERROR": logging.ERROR,
    "WARNING": logging.WARNING,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
    "NOTSET": logging.NOTSET,
}

log = logging.getLogger(__name__)
handler = None

stdout, stderr = sys.stdout, sys.stderr


def set_log_level(level: Optional[Union[int, str]] = None) -> None:
    """Set the logging level.

    If no level is passed, use INFO as logging level.
    """
    if level is None:
        level = logging.INFO
    if isinstance(level, str):
        level = log_levels[level.upper()]
    log.setLevel(level)


def set_output(stream: Optional[TextIO] = None) -> None:
    """Set the standard output.

    If no stream is passed, use standard output.
    """
    global handler
    if stream is None:
        stream = stdout
    if handler:
        log.removeHandler(handler)
    handler = logging.StreamHandler(stream)
    log.addHandler(handler)
    sys.stdout = stream


def set_err_out(stream: Optional[TextIO] = None) -> None:
    """Set the error output.

    If no stream is passed, use standard error.
    """
    if stream is None:
        stream = stderr
    sys.stderr = stream


def shutdown() -> None:
    """Shut down and flush the logging system."""
    sys.stdout.flush()
    sys.stderr.flush()
    logging.shutdown()


set_log_level()
set_output()


# initialize global dict
from . import namespaces  # noqa: E402

# a convenience function:
from .browser import browser  # noqa: E402

# the two core components of twill:
from .parse import execute_file, execute_string  # noqa: E402
from .shell import TwillCommandLoop  # noqa: E402

namespaces.init_global_dict()



"""Extension functions to help query/assert name service information.

Functions:

  * dns_resolves -- assert that a host resolves to a specific IP address.
  * dns_a -- assert that a host directly resolves to a specific IP address
  * dns_cname -- assert that a host is an alias for another hostname.
  * dnx_mx -- assert that a given host is a mail exchanger for the given name.
  * dns_ns -- assert that a given hostname is a name server for the given name.
"""

from typing import Optional

from twill.errors import TwillAssertionError

try:
    from dns.ipv4 import inet_aton
    from dns.name import from_text
    from dns.rdatatype import RdataType
    from dns.resolver import Answer, Resolver
except ImportError as error:
    msg = str(error)
    msg += "\nMust have dnspython installed to use the DNS extension module."
    raise ImportError(msg) from error


def dns_a(host: str, ipaddress: str, server: Optional[str] = None) -> None:
    """>> dns_a <name> <ipaddress> [<name server>]

    Assert that <name> resolves to <ipaddress> (and is an A record).
    Optionally use the given name server.
    """
    if not is_ip_addr(ipaddress):
        raise ValueError(
            "<ipaddress> parameter must be an IP address, not a hostname"
        )

    for answer in _resolve(host, RdataType.A, server):
        if answer.address == ipaddress:
            return

    raise TwillAssertionError


def dns_cname(host: str, cname: str, server: Optional[str] = None) -> None:
    """>> dns_cname <name> <alias_for> [<name server>]

    Assert that <name> is a CNAME alias for <alias_for>.
    Optionally use the given <name server>.
    """
    if is_ip_addr(cname):
        raise ValueError(
            "<alias_for> parameter must be a hostname, not an IP address"
        )

    cname_name = from_text(cname)

    for answer in _resolve(host, RdataType.CNAME, server):
        if answer.target == cname_name:
            return

    raise TwillAssertionError


def dns_resolves(
    host: str, ipaddress: str, server: Optional[str] = None
) -> None:
    """>> dns_resolves <name> <name2/ipaddress> [<name server>]

    Assert that <name> ultimately resolves to the given IP address (or
    the same IP address that 'name2' resolves to).
    Optionally use the given name server.
    """
    if not is_ip_addr(ipaddress):
        ipaddress = _resolve_name(ipaddress, server)

    for answer in _resolve(host, RdataType.A, server):
        if answer.address == ipaddress:
            return

    raise TwillAssertionError


def dns_mx(host: str, mailserver: str, server: Optional[str] = None) -> None:
    """>> dns_mx <name> <mailserver> [<name server>]

    Assert that <mailserver> is a mailserver for <name>.
    """
    mailserver_name = from_text(mailserver)

    for rdata in _resolve(host, RdataType.MX, server):
        if rdata.exchange == mailserver_name:
            return

    raise TwillAssertionError


def dns_ns(host: str, query_ns: str, server: Optional[str] = None) -> None:
    """>> dns_ns <domain> <nameserver> [<name server to use>]

    Assert that <nameserver> is a mailserver for <domain>.
    """
    query_ns_name = from_text(query_ns)

    for answer in _resolve(host, RdataType.NS, server):
        if answer.target == query_ns_name:
            return

    raise TwillAssertionError


def is_ip_addr(text: str) -> bool:
    """Check the 'name' to see if it's just an IP address."""
    try:
        inet_aton(text)
    except OSError:
        return False
    return True


def _resolve_name(name: str, server: Optional[str] = None) -> str:
    """Resolve the given name to an IP address."""
    if is_ip_addr(name):
        return name

    resolver = Resolver()
    if server:
        resolver.nameservers = [_resolve_name(server, None)]

    answers = resolver.resolve(name)

    return str(answers[0])


def _resolve(
    query: str, query_type: RdataType, server: Optional[str] = None
) -> Answer:
    """Resolve, perhaps via the given name server."""
    resolver = Resolver()
    if server:
        resolver.nameservers = [_resolve_name(server, None)]

    return resolver.resolve(query, query_type)



"""Extension functions for managing a SourceForge-based mailman queue.

The extension can discard all moderated messages in a mailman queue.
This is useful since currently there is no way to do this without manually
selecting 'discard' for each and every message.
"""

import re

from twill import browser, log, utils

__all__ = ["discard_all_messages", "exit_if_empty"]


def exit_if_empty() -> None:
    """>> exit_if_empty

    Exit the script currently running, if there are no deferred messages
    on the current page.
    """
    form = browser.form()
    if not form:
        log.error("No messages; exiting.")
        raise SystemExit


def discard_all_messages() -> None:
    """>> discard_all_messages

    Set all buttons to "discard".
    """
    _form_value_by_regex_setall("1", "^\\d+$", "3")


def _form_value_by_regex_setall(
    form_name: str, field_name: str, value: str
) -> None:
    form = browser.form(form_name)
    if not form:
        log.error("no such form '%s'", form_name)
        return

    regex = re.compile(field_name)

    matches = [ctl for ctl in form.inputs if regex.search(str(ctl.name))]

    if matches:
        log.info("-- matches %d", len(matches))

        n = 0
        for control in matches:
            browser.clicked(form, control)
            if "readonly" not in control.attrib:
                utils.set_form_control_value(control, value)
                n += 1

        log.info("set %d values total", n)



"""Code parsing and evaluation for the twill mini-language."""

import re
import sys
from contextlib import nullcontext
from io import StringIO
from typing import Any, Callable, Dict, List, Optional, Sequence, TextIO, Tuple

from pyparsing import (
    CharsNotIn,
    Combine,
    Group,
    Literal,
    Opt,
    ParseException,
    ParseResults,
    Word,
    ZeroOrMore,
    pyparsing_unicode,
    remove_quotes,
    rest_of_line,
)

from . import commands, log, namespaces
from .browser import browser
from .errors import TwillNameError

# pyparsing stuff

# allow characters in full 8-bit range
char_range = pyparsing_unicode.Latin1
alphas, alphanums = char_range.alphas, char_range.alphanums
printables = char_range.printables

# basically, a valid Python identifier:
command_word = Word(alphas + "_", alphanums + "_")
command = command_word.set_results_name("command")
command.set_name("command")

# arguments to it.

# we need to reimplement all this junk from pyparsing because pcre's
# idea of escapable characters contains a lot more than the C-like
# thing pyparsing implements
_bslash = "\\"
_sgl_quote = Literal("'")
_dbl_quote = Literal('"')
_escapables = printables
_escaped_char = Word(_bslash, _escapables, exact=2)
dbl_quoted_string = (
    Combine(
        _dbl_quote
        + ZeroOrMore(CharsNotIn('\\"\n\r') | _escaped_char | '""')
        + _dbl_quote
    )
    .streamline()
    .set_name("string enclosed in double quotes")
)
sgl_quoted_string = (
    Combine(
        _sgl_quote
        + ZeroOrMore(CharsNotIn("\\'\n\r") | _escaped_char | "''")
        + _sgl_quote
    )
    .streamline()
    .set_name("string enclosed in single quotes")
)
quoted_arg = dbl_quoted_string | sgl_quoted_string
quoted_arg.set_parse_action(remove_quotes)
quoted_arg.set_name("quoted_arg")

plain_arg_chars = printables.replace("#", "").replace('"', "").replace("'", "")
plain_arg = Word(plain_arg_chars)
plain_arg.set_name("plain_arg")

arguments_group = Group(ZeroOrMore(quoted_arg | plain_arg))
arguments = arguments_group.set_results_name("arguments")
arguments.set_name("arguments")

# comment line.
comment = Literal("#") + rest_of_line
comment = comment.suppress()
comment.set_name("comment")

full_command = comment | (command + arguments + Opt(comment))
full_command.set_name("full_command")


command_list: List[str] = []  # filled in by namespaces.init_global_dict().


def process_args(
    args: Sequence[str],
    globals_dict: Dict[str, Any],
    locals_dict: Dict[str, Any],
) -> List[str]:
    """Process string arguments.

    Take a list of string arguments parsed via pyparsing and evaluate
    the special variables ('__*').

    Return a new list.
    """
    new_args: List[str] = []
    for arg in args:
        # __variable substitution
        if arg.startswith("__"):
            try:
                val = eval(arg, globals_dict, locals_dict)
            except NameError:  # not in dictionary; don't interpret
                val = arg

            log.info("VAL IS %s FOR %s", val, arg)

            if isinstance(val, str):
                new_args.append(val)
            else:
                new_args.extend(val)

        # $variable substitution
        elif arg.startswith("$") and not arg.startswith("${"):
            try:
                val = str(eval(arg[1:], globals_dict, locals_dict))
            except NameError:  # not in dictionary; don't interpret
                val = arg
            new_args.append(val)
        else:
            new_args.append(
                variable_substitution(arg, globals_dict, locals_dict)
            )

    return [arg.replace("\\n", "\n") for arg in new_args]


def execute_command(
    cmd: str,
    args: Sequence[str],
    globals_dict: Dict[str, Any],
    locals_dict: Dict[str, Any],
    cmd_info: str,
) -> None:
    """Actually execute the command.

    Side effects:
     - __args__ is set to the arguments
     - __cmd__ is set to the command, and
     - __url__ is set to the browser URL.
    """
    # execute command
    locals_dict["__cmd__"] = cmd
    locals_dict["__args__"] = args
    if cmd not in command_list:
        raise TwillNameError(f"unknown twill command: '{cmd}'")

    eval_str = f"{cmd}(*__args__)"

    # compile the code object so that we can get 'cmd_info' into the
    # error tracebacks
    code_obj = compile(eval_str, cmd_info, "eval")

    # evaluate the code object in the appropriate dictionary
    result = eval(code_obj, globals_dict, locals_dict)

    # set __url__
    locals_dict["__url__"] = browser.url

    return result


_log_commands: Callable = log.debug  # type: ignore[has-type]


def parse_command(
    line: str,
    globals_dict: Dict[str, Any],
    locals_dict: Dict[str, Any],
) -> Tuple[Optional[str], Optional[List[str]]]:
    """Parse command.

    Returns a tuple with the command and its arguments.
    """
    try:
        results: Optional[ParseResults] = full_command.parse_string(line)
    except ParseException as e:
        log.error("PARSE ERROR: %s", e)
        results = None
    if results:
        _log_commands("twill: executing cmd '%s'", line.strip())
        args = process_args(
            results.arguments.as_list(), globals_dict, locals_dict
        )
        return results.command, args
    return None, None  # e.g. a comment


def execute_string(buf: str, **kw: Any) -> None:
    """Execute commands from a string buffer."""
    fp = StringIO(buf)

    kw["source"] = ["<string buffer>"]
    if "no_reset" not in kw:
        kw["no_reset"] = True

    _execute_script(fp, **kw)


def execute_file(filename: str, **kw: Any) -> None:
    """Execute commands from a file."""
    with (
        nullcontext(sys.stdin)
        if filename == "-"
        else open(filename, encoding="utf-8")
    ) as inp:
        log.info("\n>> Running twill file %s", filename)

        kw["source"] = filename
        _execute_script(inp, **kw)


def _execute_script(inp: TextIO, **kw: Any) -> None:
    """Execute lines taken from a file-like iterator."""
    # initialize new local dictionary and get global and current local
    namespaces.new_local_dict()
    globals_dict, locals_dict = namespaces.get_twill_glocals()

    locals_dict["__url__"] = browser.url

    # reset browser
    if not kw.get("no_reset"):
        commands.reset_browser()

    # go to a specific URL?
    init_url = kw.get("initial_url")
    if init_url:
        commands.go(init_url)
        locals_dict["__url__"] = browser.url

    # should we catch exceptions on failure?
    catch_errors = kw.get("never_fail")

    # source_info stuff
    source_info = kw.get("source", "<input>")

    try:
        for line_no, line_raw in enumerate(inp, 1):
            line = line_raw.strip()
            if not line:  # skip empty lines
                continue

            cmd_info = f"{source_info}:{line_no}"
            log.info("AT LINE: %s", cmd_info)

            cmd, args = parse_command(line, globals_dict, locals_dict)
            if cmd is None:
                continue

            try:
                execute_command(cmd, args, globals_dict, locals_dict, cmd_info)
            except SystemExit:
                # abort script execution if a SystemExit is raised
                return
            except Exception as error:
                error_type = error.__class__.__name__ or "Error"
                error_context = (
                    f"{error_type} raised on line {line_no} of '{source_info}'"
                )
                if line:
                    error_context += f" while executing\n>> {line}"
                if not browser.first_error:
                    browser.first_error = error_context
                log.error("\nOops! %s", error_context)
                error_msg = str(error).strip()
                log.error("\nError: %s", error_msg)
                if not catch_errors:
                    raise

    finally:
        cleanups = locals_dict.get("__cleanups__")
        if cleanups:
            first_error, result = browser.first_error, browser.result
            for filename in reversed(cleanups):
                log.info("\n>> Running twill cleanup file %s", filename)
                try:
                    with open(filename, encoding="utf-8") as cleanup:
                        _execute_script(
                            cleanup, source=filename, no_reset=True
                        )
                except Exception as error:  # noqa: BLE001
                    log.error(
                        ">> Cannot run cleanup file %s: %s", filename, error
                    )
            browser.reset()
            browser.first_error, browser.result = first_error, result
        namespaces.pop_local_dict()


def log_commands(flag: bool) -> bool:  # noqa: FBT001
    """Turn printing of commands as they are executed on or off."""
    global _log_commands
    old_flag = _log_commands is log.info
    _log_commands = log.info if flag else log.debug
    return old_flag


_re_variable = re.compile(r"\${(.*?)}")


def variable_substitution(
    raw_str: str, globals_dict: Dict[str, Any], locals_dict: Dict[str, Any]
) -> str:
    """Substitute the variables in the given string."""
    parts: List[str] = []
    append = parts.append
    position = 0
    for match in _re_variable.finditer(raw_str):
        append(raw_str[position : match.start()])
        try:
            variable = match.group(1)
            value = eval(variable, globals_dict, locals_dict)
            append(str(value))
        except NameError:
            append(match.group())
        position = match.end()
    append(raw_str[position:])
    return "".join(parts)



"""The twill exceptions."""


class TwillException(Exception):  # noqa: N818
    """General twill exception."""


class TwillAssertionError(TwillException):
    """AssertionError to raise upon failure of some twill command."""


class TwillNameError(TwillException):
    """Error to raise when an unknown command is called."""
